package sadna;

import java.awt.Frame;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.DataInputStream;
//import java.io.DataOutputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
//import java.io.FileOutputStream;
import java.io.IOException;
import java.util.List;
import java.util.Random;

import player.gamer.statemachine.sadna.FourInARow;

public class ImpFourInARow implements FourInARow {
	private String myrole;
	// 0 - empty
	// 1 - me
	// 2 - rival
	private byte field[][];
	int occurences[][]; //Row is source and column is destination. Number of occurences of possible transitions.
	float V[];
	Frame frame;
	int srcTurn; //Variable that stores last turn, made by our side (need to fill entries in delta).
	int dstTurn; //Variable that stores last turn, made by opponent (need to fill entries in delta).
	double returns[]; //Holds return from every turn (needed for second largest value calculation).
	Random generator; //Random number generator.
	boolean turnsArr[]; // Contains true if corresponding turn was made.
	double succIndicator[]; //Contains values from 0.0f to 1.0f according to value of the turn.
	boolean is_first_turn;
	boolean is_black;
	
	public ImpFourInARow(){
		field = new byte[6][8];
		occurences = new int[6561][6561]; //Occurrences matrix.
		returns = new double[8];
		turnsArr = new boolean[6561];
		succIndicator = new double[6561];
		srcTurn = dstTurn = -1;
		is_first_turn = false;
		generator = new Random();
		//Occurrences matrix initialized to 1.
		for(int i=0; i<6561; i++)
			for(int j=0; j<6561; j++)
				occurences[i][j] = 1;
		frame = new CircleDraw(field);
		frame.addWindowListener(new WindowAdapter(){
			public void windowClosing(WindowEvent we){
				System.exit(0);
			}
		});
		frame.setSize(465, 425);
		frame.setVisible(true);
	}
	@Override
	public void fourEnd(int score) {
		System.out.print(myrole +" Score: "+ score + "\n");
		field = new byte[6][8];
		/*
		//For training.
		if(score == 100)
			writePolicyFile(turnsArr,succIndicator, is_black);
		*/
	}

	@Override
	public int fourMove(List<Integer> moved, List<Integer> options, long arg2) {
		if(moved.size()>0){
			is_first_turn = false;
			ManageField.addTurn(field, moved.get(0), false);
			((CircleDraw)frame).reload(field);
		//Fills delta matrix.
		if(srcTurn == -1)
			occurences[0][ManageField.stateNumber(field)] += 1;
		else
			occurences[srcTurn][ManageField.stateNumber(field)] += 1;
		}else{
			is_first_turn = true;
		}
		double maxValue = 0.0f;
		double tempValue = 0.0f;
		int winningCol = 0;
		byte[][] tempField = new byte[6][8];
		byte[][] bestPosition = new byte[6][8];
		//Initializes returns array.
		for(int i=0; i<8; i++)
			returns[i] = 0.0f;
		
		for(int i:options){
			for(int j=0; j<6; j++)
				for(int k=0; k<8; k++)
					tempField[j][k] = field[j][k];
			tempValue = ManageField.retFromTurn(field, i);
			boolean not_to_add = false;//Defines whether we choose unchanged turn for V and succIndicator.
				for(int k=0; k<6; k++){
					if(tempField[k][i-1]==0){
						if(k>0 && tempField[k-1][i-1]==2){
							if(i>1 && i<8 && tempField[k-1][i]==0 && tempField[k-1][i-2]==0)
								not_to_add = true;
						}
						break;
					}
				}
			if(!not_to_add)
				ManageField.addTurn(tempField, i, true);
			tempValue += (double)(V[ManageField.stateNumber(tempField)]/100.0f);
			tempValue += succIndicator[ManageField.stateNumber(tempField)]/2.0d;
			returns[i-1] = tempValue;
			if(not_to_add)
				ManageField.addTurn(tempField, i, true);
			if(tempValue > maxValue){
				maxValue = tempValue;
				winningCol = i;
			}
		}
		
		//Makes random turn with probability 10%.
		//It chooses best turn, but not the very best.
		//If choice is too bad, best result is left as it was.
		//Need to comment in the case of training.
		if(generator.nextInt(100)<10 && options.size()>1){
			double goodChoise = 0.0d;
			int goodCol = 0;
			for(int i=0; i<8; i++)
				if(returns[i]<maxValue && returns[i]>goodChoise){
					goodChoise = returns[i];
					goodCol = i+1;
				}
			if((maxValue-goodChoise)<2.0d)
				winningCol = goodCol;
		}
		
		/*
		//For training.
		if(generator.nextInt(100)<=10 && options.size()>1){
				winningCol = options.get(generator.nextInt(options.size()));
		}
		*/
		for(int j=0; j<6; j++)
			for(int k=0; k<8; k++)
				bestPosition[j][k] = field[j][k];
		ManageField.addTurn(bestPosition, winningCol, true);
		if(ManageField.findFour(1, bestPosition) != 4){
			//Eliminates danger positions.
			for(int i=0; i<6; i++){
				for(int j=0; j<4; j++){
					if((field[i][j]==0 && field[i][j+1]==2 &&
							field[i][j+2]==2 && field[i][j+3]==0 && field[i][j+4]==0) ||
							(field[i][j]==0 && field[i][j+1]==0 &&
							field[i][j+2]==2 && field[i][j+3]==2 && field[i][j+4]==0)){
						if(i==0)
							if(field[i][j+1]==2)
								winningCol = j+1; //First case.
							else
								winningCol = j+5; //Second case.
						else{
							if(field[i-1][j]!=0 && field[i-1][j+1]!=0 &&
									field[i-1][j+2]!=0 && field[i-1][j+3]!=0 && field[i-1][j+4]!=0){
								if(field[i][j+1]==2)
									winningCol = j+1; //First case.
								else
									winningCol = j+5; //Second case.
							}
						}
					}
				}
			}
			for(int i=0; i<6; i++){
				for(int j=0; j<6; j++){
					if(field[i][j] == 2 && field[i][j+1] == 0 && field[i][j+2] == 2 && 
							((j>0 && field[i][j-1] != 1) || (j<5 && field[i][j+3] != 1))){
						if(i == 0)
							winningCol = j+2;
						else{
							if(field[i-1][j] != 0 && field[i-1][j+1] != 0 && field[i-1][j+2] !=0 &&
									(( j<5 && field[i-1][j+3] != 0 && field[i][j+3] != 1) || (j>0 && field[i-1][j-1] != 0 && field[i][j-1] != 1)))
								winningCol = j+2;
						}
					}	
				}
			}
		}
		//Check diagonal.
		for(int i=0; i<6; i++){
			for(int j=0; j<8; j++){
				if(i<3 && j<5 && field[i][j] == 2 && field[i+1][j+1] == 0 &&
						field[i][j+1] != 0 && field[i+2][j+2] == 2 && field[i+3][j+3] == 2)
					winningCol = j+2;
				if(i<3 && j<5 && field[i][j] == 2 && field[i+1][j+1] == 2 &&
					 field[i+2][j+2] == 0 && field[i+1][j+2] != 0 && field[i+3][j+3] == 2)
					winningCol = j+3;
			}
		}
		for(int i=0; i<6; i++){
			for(int j=7; j>=0 ;j--){
				if(i<3 && j>2 && field[i][j] == 2 && field[i+1][j-1] == 0 &&
						field[i][j-1] != 0 && field[i+2][j-2] == 2 && field[i+3][j-3] == 2)
					winningCol = j;
				if(i<3 && j>2 && field[i][j] == 2 && field[i+1][j-1] == 2 &&
						field[i+2][j-2] == 0 && field[i+1][j-2] != 0 && field[i+3][j-3] == 2)
					winningCol = j-1;
			}
		}
		/*
		//For training.
		if(is_first_turn)
			winningCol = generator.nextInt(8)+1;
		*/
		ManageField.addTurn(field, winningCol, true);
		srcTurn = ManageField.stateNumber(field); // srcTurn will contain coding of turn that was made.
		turnsArr[srcTurn] = true; //Accordingly to specification.
		((CircleDraw)frame).reload(field);
		return options.indexOf(winningCol);
	}

	@Override
	public void fourStart(String role) {
		myrole = new String(role);
		V = ManageField.valueIteration(occurences).clone();
		System.out.print("role " + role + " starting" + "\n");
		if(role.equalsIgnoreCase("black"))
			is_black = true;
		else
			is_black = false;
		readPolicyFile(succIndicator, is_black);
		//Initialize turnsArr.
		for(int i=0; i<6561; i++)
			turnsArr[i] = false;
	}	

	@Override
	public String getName() {
		// TODO Auto-generated method stub
		return "impPlayer";
	}
	
	private void readPolicyFile(double succIndicator[], boolean is_black){
		try {
			FileInputStream fin;
			if(is_black)
				fin = new FileInputStream("policy_black.dat");
			else
				fin = new FileInputStream("policy_red.dat");
			DataInputStream din = new DataInputStream(fin);
			try {
			for(int i=0; i<6561; i++)
					succIndicator[i] = din.readDouble();
			din.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	/*
	//For training.
	private void writePolicyFile(boolean turnsArr[], double succIndicator[], boolean is_black)
	{
		//Correct values of policy array.
		for(int i=0; i<6561; i++){
			if(turnsArr[i])
				succIndicator[i] += (1.0d-succIndicator[i])/2.0d;
		}

		try{
			FileOutputStream fos;
			if(is_black)
				fos = new FileOutputStream("policy_black.dat");
			else
				fos = new FileOutputStream("policy_red.dat");
			DataOutputStream dos = new DataOutputStream(fos);
			for(int i=0; i<6561; i++){
				dos.writeDouble(succIndicator[i]);
			}
			dos.close();
			}catch(Exception e){
				System.err.println("Error: " + e.getMessage());
			}
	}
	*/
	 
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		// TODO Auto-generated method stub

	}

}
