package sadna;
public class ManageField {
	private enum Dir {NE,E,SE,S,SW,W,NW};
	//Adds cell(1 is a first column).
	public static void addTurn(byte field[][], int colNumber, boolean isMe){
		for(int i=0; i<6; i++){
			if(field[i][colNumber-1] == 0){
				if(isMe)
					field[i][colNumber-1] = 1;
				else
					field[i][colNumber-1] = 2;
				break;
			}
		}
	}
	//Returns mark of a given turn.
	public static int retFromTurn(byte[][] field, int colNumber){
		byte[][] tempField = new byte[6][8];
		for(int i=0; i<6; i++){
			for(int j=0; j<8; j++){
				tempField[i][j] = field[i][j];
			}
		}
		addTurn(tempField, colNumber, true);
		int myReturn = maxReturn(tempField, 1);
		int rivalReturn = maxReturn(tempField,2);
		if(rivalReturn > 2 && myReturn < 4)
			rivalReturn *= 3;
		return 12+2*myReturn-rivalReturn;
	}
	//Here state sequential number is counted (in a base of three),
	//accordingly to the next list : 0-empty, 1-our mark, 2-reival's mark,
	//from left to right. Function returns the number of state.
	public static int stateNumber(byte[][] field)
	{
		int value = 0;
		for(int i=0; i<8; i++){
			for(int j=0; j<6; j++){
				if(field[j][i] == 0){
					if(j>0)
						value += field[j-1][i]*(int)Math.pow(3,i);
					break;
				}
			}
		}
		return value;
	}
	//Gets matrix of transitions probabilities (rows source state, column destination state).
	public static float[] valueIteration(int[][]  occurences)
	{
		float[] V = new float[6561];
		float d = 0.0f;
		int returnFromTurn = 0;
		float resultingValue = 0.0f;
		float sum = 0.0f;
		float gamma = 0.1f; // Parameter.
		do
		{
		d = 0.0f;
		for(int s=0; s<6561; s++){
			//System.out.print(s+" ");
			float v = V[s];
			resultingValue = 0.0f;
			//Iterate over turns.
			for(int j=0; j<8; j++){
				sum = 0.0f;
				byte[][] tempField = new byte[6][8];
				int tempVal = s;
				for(int k=0; k<8; k++){
					tempField[0][k] = (byte)(tempVal%3);
					tempVal = tempVal/3;
					if(tempVal == 0)
						break;
				}
				
					returnFromTurn = ManageField.retFromTurn(tempField, j+1);
					ManageField.addTurn(tempField, j+1, true);
					
					//Iterate over rival turns.
					//Sums occurrences of every possible turn.
					int sum_of_occ = 0;
					for(int k=0; k<8; k++){
						byte[][] tempFieldTrial = new byte[6][8];
						for(int l=0; l<6; l++)
							for(int m=0; m<8; m++)
								tempFieldTrial[l][m] = tempField[l][m];
									sum_of_occ+=occurences[ManageField.stateNumber(tempField)][ManageField.stateNumber(tempFieldTrial)];
					}
					for(int k=0; k<8; k++){
						byte[][] tempFieldTrial = new byte[6][8];
						for(int l=0; l<6; l++)
							for(int m=0; m<8; m++)
								tempFieldTrial[l][m] = tempField[l][m];
						ManageField.addTurn(tempFieldTrial, k+1, false);
						//Second strung,((float) ...) estimates probability of a turn, based on number of occurrences of such a turn.
						sum += gamma*
								((float)occurences[ManageField.stateNumber(tempField)][ManageField.stateNumber(tempFieldTrial)]/(float)sum_of_occ)*
								V[ManageField.stateNumber(tempFieldTrial)];
					}
					
					if((returnFromTurn + sum) > resultingValue)
						resultingValue = returnFromTurn + sum;
			}
			V[s] = resultingValue;
			d = Math.max(d, Math.abs(v-V[s]));
			//System.out.println("d:"+d + " sum:"+sum+" V[s]:"+V[s]);
		}
		}while(d > 1.0f);
		return V;
	}
	
	//Returns lookFor's score in a given state.
	public static int maxReturn(byte[][] field, int lookFor){
		int[] returns = new int[7];
		int maxReturn = 0;
		for(int i=0; i<8; i++){
			for(int j=0; j<6; j++){
				if(field[j][i] == 0){
					if(j+1>=0 && i-1>=0 && j+1<=5 && i-1<=7)
						returns[0] = directionReturn(field, 0, Dir.NE, lookFor, j+1, i-1);
					if(j>=0 && i-1>=0 && j<=5 && i-1<=7)
						returns[1] = directionReturn(field, 0, Dir.E, lookFor, j, i-1);
					if(j-1>=0 && i-1>=0 && j-1<=5 && i-1<=7)
						returns[2] = directionReturn(field, 0, Dir.SE, lookFor, j-1, i-1);
					if(j-1>=0 && i>=0 && j-1<=5 && i<=7){
						//System.out.println("Call for S"+"row:"+(j-1)+"col:"+i);
						returns[3] = directionReturn(field, 0, Dir.S, lookFor, j-1, i);
						//System.out.println("Called for S return:"+returns[3]);
					}
					if(j-1>=0 && i+1>=0 && j-1<=5 && i+1<=7)
						returns[4] = directionReturn(field, 0, Dir.SW, lookFor, j-1, i+1);
					if(j>=0 && i+1>=0 && j<=5 && i+1<=7)
						returns[5] = directionReturn(field, 0, Dir.W, lookFor, j, i+1);
					if(j+1>=0 && i+1>=0 && j+1<=5 && i+1<=7)
						returns[6] = directionReturn(field, 0, Dir.NW, lookFor, j+1, i+1);
					break;
				}
			}
			
			for(int k=0; k<7; k++){
				if(returns[k]>maxReturn)
					maxReturn = returns[k];
			}
		}
		
		if(findFour(lookFor, field) == 4)
			return 4;
		
		return maxReturn;
	}
	//Finds winning position and returns 4 if found, otherwise returns 0.
	public static int findFour(int lookFor, byte[][] field){
		//Look for four points in a rows.
				int counter = 0;
				for(int i=0; i<6; i++){
					for(int j=0; j<8; j++){
						if(field[i][j] == lookFor){
							counter++;
							if(counter == 4)
								return 4;
						}
						else
							counter = 0;
					}
					counter = 0;
				}
				
				counter = 0;
				
				//Look for four points in a columns.
				for(int i=0; i<8; i++){
					for(int j=0; j<6; j++){
						if(field[j][i] == lookFor){
							counter++;
							if(counter == 4)
								return 4;
						}
						else
							counter = 0;
					}
					counter = 0;
				}
				
				counter = 0;
				
				//Look for four points in a diagonal NW.
				for(int i=-5; i<8; i++){
					for(int j=0; j<6; j++){
						if(j>=0 && j<=7 && j+i>=0 && j+i<=7)
							if(field[j][j+i] == lookFor){
								counter++;
								if(counter == 4)
									return 4;
							}
							else
								counter = 0;	
					}
					counter = 0;
				}
				
				counter = 0;
				
				//Look for four points in a diagonal NE.
				for(int i=12; i>=0; i--){
					for(int j=0; j<6; j++){
						if(j>=0 && j<=7 && i-j>=0 && i-j<=7)
							if(field[j][i-j] == lookFor){
								counter++;
								if(counter == 4)
									return 4;
							}
							else
								counter = 0;	
					}
					counter = 0;
				}
				return 0;
	}
	
	private static int directionReturn(byte[][] field, int accPoints, 
			Dir direction, int lookFor, int row, int col){
		
		if(direction == Dir.NE){
			if(field[row][col] == lookFor){
				if(row+1>=0 && col-1>=0 && row+1<=5 && col-1<=7){
					return directionReturn(field, accPoints+1, direction,
								lookFor, row+1, col-1);
				}else{
					return accPoints+1;
				}
			}else
				return accPoints;
		}else if(direction == Dir.E){
			if(field[row][col] == lookFor){
				if(row>=0 && col-1>=0 && row<=5 && col-1<=7){
					return directionReturn(field, accPoints+1, direction,
								lookFor, row, col-1);
				}else{
					return accPoints+1;
				}
			}else
				return accPoints;
		}else if(direction == Dir.SE){
			if(field[row][col] == lookFor){
				if(row-1>=0 && col-1>=0 && row-1<=5 && col-1<=7){
					return directionReturn(field, accPoints+1, direction,
								lookFor, row-1, col-1);
				}else{
					return accPoints+1;
				}
			}else
				return accPoints;
		}else if(direction == Dir.S){
			if(field[row][col] == lookFor){
				//System.out.println("Search in south direction"+
				//		accPoints+"lookFor:"+lookFor+"row:"+row+"col:"+col);
				if(row-1>=0 && col>=0 && row-1<=5 && col<=7){
					return directionReturn(field, accPoints+1, direction,
								lookFor, row-1, col);
				}else{
					return accPoints+1;
				}
			}else
				return accPoints;
		}else if(direction == Dir.SW){
			if(field[row][col] == lookFor){
				if(row-1>=0 && col+1>=0 && row-1<=5 && col+1<=7){
					return directionReturn(field, accPoints+1, direction,
								lookFor, row-1, col+1);
				}else{
					return accPoints+1;
				}
			}else
				return accPoints;
		}else if(direction == Dir.W){
			if(field[row][col] == lookFor){
				if(row>=0 && col+1>=0 && row<=5 && col+1<=7){
					return directionReturn(field, accPoints+1, direction,
								lookFor, row, col+1);
				}else{
					return accPoints+1;
				}
			}else
				return accPoints;
		}else if(direction == Dir.NW){
			if(field[row][col] == lookFor){
				if(row+1>=0 && col+1>=0 && row+1<=5 && col+1<=7){
					return directionReturn(field, accPoints+1, direction,
								lookFor, row+1, col+1);
				}else{
					return accPoints+1;
				}
			}else
				return accPoints;
		}
		return 0;
	}
}
